==> TopdataTopFeedSW6-v7.1.1.zip: TopdataTopFeedSW6/src/Controller/TopFeedApiController.php <==
<?php declare(strict_types=1);

namespace Topdata\TopdataTopFeedSW6\Controller;

use Doctrine\DBAL\Connection;
use Shopware\Core\Content\Product\ProductCollection;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\NotFilter;
use Shopware\Core\Framework\Uuid\Uuid;
use Shopware\Core\System\SalesChannel\SalesChannelContext;
use Shopware\Storefront\Controller\StorefrontController;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Routing\Annotation\Route;
use Topdata\TopdataFoundationSW6\Util\UtilJsonResponse;
use Topdata\TopdataTopFeedSW6\Service\LinkedProductService;
use Topdata\TopdataTopFeedSW6\Service\TopfeedHelperServiceV1;
use Topdata\TopdataTopFeedSW6\Service\SettingsService;
use Topdata\TopdataTopFeedSW6\Storefront\Page\CompatibleDevicesWidget\CompatibleDevicesWidgetPageV1Loader;
use Topdata\TopdataTopFeedSW6\Storefront\Page\LinkedProductsPopup\LinkedProductsPopupLoader;
use Topdata\TopdataTopFeedSW6\Storefront\Page\LinkedProductsPopup\LinkTypeEnum;
use Topdata\TopdataTopFeedSW6\Storefront\Page\LinkedProductsPopup\ProductIdRetriever;

#[Route(defaults: ['_routeScope' => ['storefront']])]
class TopFeedApiController extends StorefrontController
{


    public function __construct(
        private readonly LinkedProductsPopupLoader           $linkedProductsPopupLoader,
        private readonly SettingsService                     $settingsService,
        private readonly CompatibleDevicesWidgetPageV1Loader $CompatibleDevicesWidgetPageLoader,
        private readonly ProductIdRetriever                  $productIdRetriever,
        private readonly TopfeedHelperServiceV1                $topfeedHelperServiceV1,
        private readonly LinkedProductService                $linkedProductService
    )
    {
    }

    /**
     * Retrieves linked products for a given product and link type.
     *
     * This method is used to fetch and render linked products in a popup.
     * It's typically called via AJAX to dynamically load related product information.
     *
     * @param Request $request The incoming request object
     * @param SalesChannelContext $salesChannelContext The current sales channel context
     * @return JsonResponse A JSON response containing the rendered HTML for the linked products
     */
    #[Route(
        path: '/topdata-top-feed/products/{linkType}/{productId}',
        name: 'frontend.topdata_top_feed.products',
        defaults: ['XmlHttpRequest' => true],
        methods: ['GET'])
    ]
    public function productsLinked($linkType, $productId, Request $request, SalesChannelContext $salesChannelContext): JsonResponse
    {
        if (!Uuid::isValid($productId)) {
            return UtilJsonResponse::error('Product id invalid');
        }

        $result = ['success' => true];
        $page = $this->linkedProductsPopupLoader->buildLinkedProductsPopupPage(
            $request,
            $salesChannelContext,
            $productId,
            $linkType
        );

        $body = $this->renderStorefront('@TopdataTopFeedSW6/storefront/page/linked-products-popup/index.html.twig', [
            'page' => $page
        ])->getContent();
        $result['body'] = $body;
        $result['title'] = $page->pageTitle;

        $response = new JsonResponse($result);
        $response->headers->set("X-Robots-Tag", "noindex, nofollow");

        return $response;
    }


    /**
     * TODO: linkType and $productId should be passed as parameters
     *
     * Retrieves all linked products for a given product.
     *
     * This method is used to fetch and render all linked products without pagination.
     * It's typically used when displaying a full list of variants or related products.
     *
     * @param Request $request The incoming request object
     * @param SalesChannelContext $salesChannelContext The current sales channel context
     * @return JsonResponse A JSON response containing the rendered HTML for all linked products
     */
    #[Route(
        path: '/topdata-top-feed/list-linked-products-all',
        name: 'frontend.topdata_top_feed.list_linked_products_all',
        defaults: ['XmlHttpRequest' => true],
        methods: ['GET']
    )]
    public function listLinkedProductsAll(Request $request, SalesChannelContext $salesChannelContext): JsonResponse
    {
        $productId = $request->get('productId');
        $linkType = $request->get('linkType');
        $linkTypeEnum = LinkTypeEnum::from($linkType);
        $data = '';
        $products = [];
        $products['variants'] = $this->linkedProductService->getLinkedProducts(
            $linkTypeEnum,
            $productId,
            $salesChannelContext,
            false
        );

        foreach ($products as $key => $value) {
            if (!$value['entities']) {
                unset($products[$key]);
                continue;
            }

            if (
                $linkTypeEnum === LinkTypeEnum::COLOR_VARIANTS
                || $linkTypeEnum === LinkTypeEnum::CAPACITY_VARIANTS
            ) {
                $linkedProductIds = $value['entities']->getIds();
                if (count($linkedProductIds) === 0) {
                    continue;
                }

                $products[$key]['otherDevices'] = $this->topfeedHelperServiceV1->getProductIdsWithOtherDevices(
                    $this->container->get(Connection::class),
                    $productId,
                    $linkedProductIds
                );
            }
        }

        if ($products !== []) {
            $data = $this->renderStorefront('@TopdataTopFeedSW6/storefront/page/list-linked-products/all-variants.html.twig', [
                'products' => $products
            ])->getContent();
        }
        $response = new JsonResponse(['success' => true, 'html' => $data]);
        $response->headers->set("X-Robots-Tag", "noindex, nofollow");
        return $response;
    }


    /**
     * Retrieves compatible devices for a given product.
     *
     * This method is used to fetch and render a list of devices compatible with the specified product.
     * It's typically called via AJAX to dynamically load compatibility information.
     *
     * @param string $productid The ID of the product to find compatible devices for
     * @param Request $request The incoming request object
     * @param SalesChannelContext $salesChannelContext The current sales channel context
     * @return JsonResponse A JSON response containing the rendered HTML for compatible devices
     */
    #[Route(
        path: '/topdata-top-feed/compatible-devices/{productid}',
        name: 'frontend.topdata_top_feed.compatible_devices',
        defaults: ['productid' => '', 'XmlHttpRequest' => true],
        methods: ['GET', 'POST']
    )]
    public function compatibleDevices(string $productid, Request $request, SalesChannelContext $salesChannelContext): JsonResponse
    {
        if (!$productid || !Uuid::isValid($productid)) {
            return UtilJsonResponse::error('Something went wrong');
        }

        $result = ['success' => true];
        $page = $this->CompatibleDevicesWidgetPageLoader->load($request, $salesChannelContext, $productid);
        $result['html'] = $this->renderStorefront('@TopdataTopFeedSW6/storefront/page/product-detail/feed-devices-content.html.twig', [
            'page' => $page
        ])->getContent();

        $response = new JsonResponse($result);
        $response->headers->set("X-Robots-Tag", "noindex, nofollow");
        return $response;
    }


    /**
     * TODO: move parts of this to LinkedProductService
     *
     * Retrieves an optimized list of linked products for multiple products.
     *
     * This method is used to fetch and render linked products for multiple products in an optimized manner.
     * It's designed to handle bulk requests efficiently, reducing database queries and improving performance.
     *
     * @param Request $request The incoming request object
     * @param SalesChannelContext $salesChannelContext The current sales channel context
     * @return JsonResponse A JSON response containing the rendered HTML for linked products of multiple products
     */
    #[Route(
        path: '/topdata-top-feed/list-linked-products',
        name: 'frontend.topdata_top_feed.list_linked_products',
        defaults: ['XmlHttpRequest' => true],
        methods: ['GET', 'POST']
    )]
    public function listLinkedProductsOptimized(Request $request, SalesChannelContext $salesChannelContext): JsonResponse
    {
        $limit = $this->settingsService->getInt('listVariantsLimit');
        $productIds = $request->get('ids');
        $data = [];
        $linkedProducts = [];
        $allLinkedProdIds = [];

        if ($this->settingsService->getBool('listHideForVariated') && is_array($productIds)) {
            $realIds = [];
            foreach ($productIds as $id) {
                if (Uuid::isValid($id)) {
                    $realIds[] = $id;
                }
            }
            $realIds = array_unique($realIds);
            $criteria = new Criteria($realIds);
            $criteria->addFilter(
                new NotFilter(NotFilter::CONNECTION_OR, [new EqualsFilter('parentId', null)])
            );
            $productVariations = $this
                ->container
                ->get('sales_channel.product.repository')
                ->search(
                    $criteria,
                    $salesChannelContext
                )
                ->getEntities();
        } else {
            $productVariations = new ProductCollection();
        }

        foreach ($productIds as $id) {
            $products = [];

            foreach ($this->settingsService->getLinkedSettings() as $key => $enabled) {
                $linkTypeEnum = LinkTypeEnum::from($key);
                if (($linkTypeEnum === LinkTypeEnum::COLOR_VARIANTS
                        ||
                        $linkTypeEnum === LinkTypeEnum::CAPACITY_VARIANTS
                    )
                    && $productVariations->get($id)
                ) {
                    continue;
                }
                if ($enabled) {
                    $ids = $this->productIdRetriever->getLinkedProductIdsByLinkType($linkTypeEnum, $id);
                    if (count($ids) === 0) {
                        continue;
                    }
                    $hasMore = $limit && (count($ids) > $limit);
                    if ($limit !== 0) {
                        $ids = array_slice($ids, 0, $limit);
                    }
                    $products[$key] = [
                        'ids'          => $ids,
                        'entities'     => [],
                        'otherDevices' => [],
                        'hasMore'      => $hasMore
                    ];
                    $allLinkedProdIds = array_merge($allLinkedProdIds, $ids);
                }
            }

            foreach ($products as $key => $value) {
                if (
                    $key === LinkTypeEnum::COLOR_VARIANTS->value
                    || $key === LinkTypeEnum::CAPACITY_VARIANTS->value
                ) {
                    $products[$key]['otherDevices'] = $this->topfeedHelperServiceV1->getProductIdsWithOtherDevices(
                        $id,
                        $value['ids']
                    );
                }
            }

            if ($products !== []) {
                $linkedProducts[$id] = $products;
            }
        }

        $allLinkedProdIds = array_unique($allLinkedProdIds);

        if ($allLinkedProdIds !== []) {
            $allLinkedProducts = $this
                ->container
                ->get('sales_channel.product.repository')
                ->search(
                    (new Criteria($allLinkedProdIds))->addAssociations(['prices', 'properties.group', 'manufacturer']),
                    $salesChannelContext
                )
                ->getEntities();
        } else {
            $allLinkedProducts = new ProductCollection();
        }


        foreach ($linkedProducts as $id => $linkedInfo) {
            foreach ($linkedInfo as $linkType => $info) {
                $prods = [];
                foreach ($info['ids'] as $lid) {
                    if ($allLinkedProducts->get($lid)) {
                        $prods[] = $allLinkedProducts->get($lid);
                    }
                }
                if ($prods !== []) {
                    $linkedProducts[$id][$linkType]['entities'] = new ProductCollection($prods);
                }
            }

        }

        foreach ($linkedProducts as $id => $linkedInfo) {
            $data[] = [
                'id'   => $id,
                'html' => $this->renderStorefront('@TopdataTopFeedSW6/storefront/page/list-linked-products/index.html.twig', [
                    'products'        => $linkedInfo,
                    'parentProductId' => $id
                ])->getContent()
            ];
        }

        return new JsonResponse(['success' => true, 'products' => $data]);
    }


}

==> TopdataTopFeedSW6-v7.1.2.zip: TopdataTopFeedSW6/src/Controller/TopFeedApiController.php <==
<?php declare(strict_types=1);

namespace Topdata\TopdataTopFeedSW6\Controller;

use Doctrine\DBAL\Connection;
use Shopware\Core\Content\Product\ProductCollection;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\NotFilter;
use Shopware\Core\Framework\Uuid\Uuid;
use Shopware\Core\System\SalesChannel\SalesChannelContext;
use Shopware\Storefront\Controller\StorefrontController;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Routing\Annotation\Route;
use Topdata\TopdataTopFeedSW6\Foundation\Util\UtilJsonResponse;
use Topdata\TopdataTopFeedSW6\Service\LinkedProductService;
use Topdata\TopdataTopFeedSW6\Service\TopfeedHelperServiceV1;
use Topdata\TopdataTopFeedSW6\Service\SettingsService;
use Topdata\TopdataTopFeedSW6\Storefront\Page\CompatibleDevicesWidget\CompatibleDevicesWidgetPageV1Loader;
use Topdata\TopdataTopFeedSW6\Storefront\Page\LinkedProductsPopup\LinkedProductsPopupLoader;
use Topdata\TopdataTopFeedSW6\Storefront\Page\LinkedProductsPopup\LinkTypeEnum;
use Topdata\TopdataTopFeedSW6\Storefront\Page\LinkedProductsPopup\ProductIdRetriever;

#[Route(defaults: ['_routeScope' => ['storefront']])]
class TopFeedApiController extends StorefrontController
{


    public function __construct(
        private readonly LinkedProductsPopupLoader           $linkedProductsPopupLoader,
        private readonly SettingsService                     $settingsService,
        private readonly CompatibleDevicesWidgetPageV1Loader $CompatibleDevicesWidgetPageLoader,
        private readonly ProductIdRetriever                  $productIdRetriever,
        private readonly TopfeedHelperServiceV1                $topfeedHelperServiceV1,
        private readonly LinkedProductService                $linkedProductService
    )
    {
    }

    /**
     * Retrieves linked products for a given product and link type.
     *
     * This method is used to fetch and render linked products in a popup.
     * It's typically called via AJAX to dynamically load related product information.
     *
     * @param Request $request The incoming request object
     * @param SalesChannelContext $salesChannelContext The current sales channel context
     * @return JsonResponse A JSON response containing the rendered HTML for the linked products
     */
    #[Route(
        path: '/topdata-top-feed/products/{linkType}/{productId}',
        name: 'frontend.topdata_top_feed.products',
        defaults: ['XmlHttpRequest' => true],
        methods: ['GET'])
    ]
    public function productsLinked($linkType, $productId, Request $request, SalesChannelContext $salesChannelContext): JsonResponse
    {
        if (!Uuid::isValid($productId)) {
            return UtilJsonResponse::error('Product id invalid');
        }

        $result = ['success' => true];
        $page = $this->linkedProductsPopupLoader->buildLinkedProductsPopupPage(
            $request,
            $salesChannelContext,
            $productId,
            $linkType
        );

        $body = $this->renderStorefront('@TopdataTopFeedSW6/storefront/page/linked-products-popup/index.html.twig', [
            'page' => $page
        ])->getContent();
        $result['body'] = $body;
        $result['title'] = $page->pageTitle;

        $response = new JsonResponse($result);
        $response->headers->set("X-Robots-Tag", "noindex, nofollow");

        return $response;
    }


    /**
     * TODO: linkType and $productId should be passed as parameters
     *
     * Retrieves all linked products for a given product.
     *
     * This method is used to fetch and render all linked products without pagination.
     * It's typically used when displaying a full list of variants or related products.
     *
     * @param Request $request The incoming request object
     * @param SalesChannelContext $salesChannelContext The current sales channel context
     * @return JsonResponse A JSON response containing the rendered HTML for all linked products
     */
    #[Route(
        path: '/topdata-top-feed/list-linked-products-all',
        name: 'frontend.topdata_top_feed.list_linked_products_all',
        defaults: ['XmlHttpRequest' => true],
        methods: ['GET']
    )]
    public function listLinkedProductsAll(Request $request, SalesChannelContext $salesChannelContext): JsonResponse
    {
        $productId = $request->get('productId');
        $linkType = $request->get('linkType');
        $linkTypeEnum = LinkTypeEnum::from($linkType);
        $data = '';
        $products = [];
        $products['variants'] = $this->linkedProductService->getLinkedProducts(
            $linkTypeEnum,
            $productId,
            $salesChannelContext,
            false
        );

        foreach ($products as $key => $value) {
            if (!$value['entities']) {
                unset($products[$key]);
                continue;
            }

            if (
                $linkTypeEnum === LinkTypeEnum::COLOR_VARIANTS
                || $linkTypeEnum === LinkTypeEnum::CAPACITY_VARIANTS
            ) {
                $linkedProductIds = $value['entities']->getIds();
                if (count($linkedProductIds) === 0) {
                    continue;
                }

                $products[$key]['otherDevices'] = $this->topfeedHelperServiceV1->getProductIdsWithOtherDevices(
                    $this->container->get(Connection::class),
                    $productId,
                    $linkedProductIds
                );
            }
        }

        if ($products !== []) {
            $data = $this->renderStorefront('@TopdataTopFeedSW6/storefront/page/list-linked-products/all-variants.html.twig', [
                'products' => $products
            ])->getContent();
        }
        $response = new JsonResponse(['success' => true, 'html' => $data]);
        $response->headers->set("X-Robots-Tag", "noindex, nofollow");
        return $response;
    }


    /**
     * Retrieves compatible devices for a given product.
     *
     * This method is used to fetch and render a list of devices compatible with the specified product.
     * It's typically called via AJAX to dynamically load compatibility information.
     *
     * @param string $productid The ID of the product to find compatible devices for
     * @param Request $request The incoming request object
     * @param SalesChannelContext $salesChannelContext The current sales channel context
     * @return JsonResponse A JSON response containing the rendered HTML for compatible devices
     */
    #[Route(
        path: '/topdata-top-feed/compatible-devices/{productid}',
        name: 'frontend.topdata_top_feed.compatible_devices',
        defaults: ['productid' => '', 'XmlHttpRequest' => true],
        methods: ['GET', 'POST']
    )]
    public function compatibleDevices(string $productid, Request $request, SalesChannelContext $salesChannelContext): JsonResponse
    {
        if (!$productid || !Uuid::isValid($productid)) {
            return UtilJsonResponse::error('Something went wrong');
        }

        $result = ['success' => true];
        $page = $this->CompatibleDevicesWidgetPageLoader->load($request, $salesChannelContext, $productid);
        $result['html'] = $this->renderStorefront('@TopdataTopFeedSW6/storefront/page/product-detail/feed-devices-content.html.twig', [
            'page' => $page
        ])->getContent();

        $response = new JsonResponse($result);
        $response->headers->set("X-Robots-Tag", "noindex, nofollow");
        return $response;
    }


    /**
     * TODO: move parts of this to LinkedProductService
     *
     * Retrieves an optimized list of linked products for multiple products.
     *
     * This method is used to fetch and render linked products for multiple products in an optimized manner.
     * It's designed to handle bulk requests efficiently, reducing database queries and improving performance.
     *
     * @param Request $request The incoming request object
     * @param SalesChannelContext $salesChannelContext The current sales channel context
     * @return JsonResponse A JSON response containing the rendered HTML for linked products of multiple products
     */
    #[Route(
        path: '/topdata-top-feed/list-linked-products',
        name: 'frontend.topdata_top_feed.list_linked_products',
        defaults: ['XmlHttpRequest' => true],
        methods: ['GET', 'POST']
    )]
    public function listLinkedProductsOptimized(Request $request, SalesChannelContext $salesChannelContext): JsonResponse
    {
        $limit = $this->settingsService->getInt('listVariantsLimit');
        $productIds = $request->get('ids');
        $data = [];
        $linkedProducts = [];
        $allLinkedProdIds = [];

        if ($this->settingsService->getBool('listHideForVariated') && is_array($productIds)) {
            $realIds = [];
            foreach ($productIds as $id) {
                if (Uuid::isValid($id)) {
                    $realIds[] = $id;
                }
            }
            $realIds = array_unique($realIds);
            $criteria = new Criteria($realIds);
            $criteria->addFilter(
                new NotFilter(NotFilter::CONNECTION_OR, [new EqualsFilter('parentId', null)])
            );
            $productVariations = $this
                ->container
                ->get('sales_channel.product.repository')
                ->search(
                    $criteria,
                    $salesChannelContext
                )
                ->getEntities();
        } else {
            $productVariations = new ProductCollection();
        }

        foreach ($productIds as $id) {
            $products = [];

            foreach ($this->settingsService->getLinkedSettings() as $key => $enabled) {
                $linkTypeEnum = LinkTypeEnum::from($key);
                if (($linkTypeEnum === LinkTypeEnum::COLOR_VARIANTS
                        ||
                        $linkTypeEnum === LinkTypeEnum::CAPACITY_VARIANTS
                    )
                    && $productVariations->get($id)
                ) {
                    continue;
                }
                if ($enabled) {
                    $ids = $this->productIdRetriever->getLinkedProductIdsByLinkType($linkTypeEnum, $id);
                    if (count($ids) === 0) {
                        continue;
                    }
                    $hasMore = $limit && (count($ids) > $limit);
                    if ($limit !== 0) {
                        $ids = array_slice($ids, 0, $limit);
                    }
                    $products[$key] = [
                        'ids'          => $ids,
                        'entities'     => [],
                        'otherDevices' => [],
                        'hasMore'      => $hasMore
                    ];
                    $allLinkedProdIds = array_merge($allLinkedProdIds, $ids);
                }
            }

            foreach ($products as $key => $value) {
                if (
                    $key === LinkTypeEnum::COLOR_VARIANTS->value
                    || $key === LinkTypeEnum::CAPACITY_VARIANTS->value
                ) {
                    $products[$key]['otherDevices'] = $this->topfeedHelperServiceV1->getProductIdsWithOtherDevices(
                        $id,
                        $value['ids']
                    );
                }
            }

            if ($products !== []) {
                $linkedProducts[$id] = $products;
            }
        }

        $allLinkedProdIds = array_unique($allLinkedProdIds);

        if ($allLinkedProdIds !== []) {
            $allLinkedProducts = $this
                ->container
                ->get('sales_channel.product.repository')
                ->search(
                    (new Criteria($allLinkedProdIds))->addAssociations(['prices', 'properties.group', 'manufacturer']),
                    $salesChannelContext
                )
                ->getEntities();
        } else {
            $allLinkedProducts = new ProductCollection();
        }


        foreach ($linkedProducts as $id => $linkedInfo) {
            foreach ($linkedInfo as $linkType => $info) {
                $prods = [];
                foreach ($info['ids'] as $lid) {
                    if ($allLinkedProducts->get($lid)) {
                        $prods[] = $allLinkedProducts->get($lid);
                    }
                }
                if ($prods !== []) {
                    $linkedProducts[$id][$linkType]['entities'] = new ProductCollection($prods);
                }
            }

        }

        foreach ($linkedProducts as $id => $linkedInfo) {
            $data[] = [
                'id'   => $id,
                'html' => $this->renderStorefront('@TopdataTopFeedSW6/storefront/page/list-linked-products/index.html.twig', [
                    'products'        => $linkedInfo,
                    'parentProductId' => $id
                ])->getContent()
            ];
        }

        return new JsonResponse(['success' => true, 'products' => $data]);
    }


}
